#' do atom to seq alignment
#' 
#' @param atomSeq AAString sequence
#' @param seqSeq AAString sequence
#' @return list of results, most important is atomNum, which is the numbering to make ATOM match the SEQ
#' @importFrom Biostrings pairwiseAlignment insertion deletion
#' @import IRanges
#' @export
genAlign <- function(atomSeq, seqSeq){
  if ((length(atomSeq) == 0) || (length(seqSeq) == 0)){
    return(list(seqRanges=NULL, atomRanges=NULL, atomNum=NULL, didAlign=FALSE, hasInsertionDeletion=FALSE))
  }
  
  outAlign <- pairwiseAlignment(atomSeq, seqSeq)
  seqRanges <- IRanges(start=start(subject(outAlign)), end=end(subject(outAlign)))
  atomRanges <- IRanges(start=start(pattern(outAlign)), width=nchar(pattern(outAlign)))
  
  # figuring out the "real" insertions, deletions
  insertions <- insertion(outAlign)[[1]]
  deletions <- deletion(outAlign)[[1]]
  
  tmpAtomNum <- rep(0, length(atomSeq))
  
  # create the intervals, and annotate them as to their type
  allIntervals <- disjoin(c(atomRanges, insertions, deletions))
  intervalType <- rep("real", length(allIntervals))
  suppressWarnings(intervalType[allIntervals %in% insertions] <- "insertion")
  suppressWarnings(intervalType[allIntervals %in% deletions] <- "deletion")
  
  seqIndex <- start(seqRanges)
  atomIndex <- start(atomRanges[1])
  
  for (iInterval in seq(1, length(allIntervals))){
    useWidth <- width(allIntervals[iInterval])
    switch(intervalType[iInterval],
           real={
             useSeq <- seq(seqIndex, seqIndex + useWidth - 1)
             useIndex <- seq(atomIndex, atomIndex + useWidth - 1)
             tmpAtomNum[useIndex] <- useSeq
             seqIndex <- seqIndex + useWidth
             atomIndex <- atomIndex + useWidth
           },
           insertion={
             atomIndex <- atomIndex + useWidth
           },
           deletion={
             seqIndex <- seqIndex + useWidth
           })
  }
  
  didAlign <- TRUE
  
  return(list(seqRanges=seqRanges, atomRanges=atomRanges, atomNum=tmpAtomNum, didAlign=TRUE))
}

#' parse ligand info
#' 
#' parses out the ligand information from the fasta sequence headers. Assumes that header is:
#' PDBID.CHAIN.LENGTH|CHAIN-NUM.LIGANDMOL.ATOMID.ATOMTYPE|...|...|...|CHAINSTART
#' 
#' @param header the fasta sequence header to parse
#' @param ligandSep what separates ligands from each other (defaults to "|")
#' @param innerSep what separates identifiers within a ligand (defaults to ".")
#' @return data.frame of locs and ligands
#' @export
parseFHeader <- function(header, ligandSep="|", innerSep="."){
  stopifnot(nchar(header) > 0)
  
  splitHeader <- strsplit(header, ligandSep, fixed=TRUE)[[1]]
  nSplit <- length(splitHeader)
  splitHeader <- splitHeader[2:(nSplit-1)]
  innerSplit <- strsplit(splitHeader, innerSep, fixed=TRUE)
  
  outLigand <- lapply(innerSplit, function(x){
    loc <- as.numeric(substring(x[1], 2))
    ligand <- x[2]
    ligandShort <- "-"
    matchLigand <- which(toupper(AMINO_ACID_CODE) %in% ligand)
    if (length(matchLigand) != 0){
      ligandShort <- names(AMINO_ACID_CODE)[matchLigand]
    }
    data.frame(loc=loc, ligand=ligand, lShort=ligandShort, atom=x[3], element=x[4], stringsAsFactors=FALSE)
  })
  outLigand <- do.call(rbind, outLigand)
  return(outLigand)
}

#' parses ATOM start from header
#' 
#' Frequently the ATOM records do not start with a 1, but some other number. This number is encoded
#' at the end of the header record.
#' 
#' @param header the header to parse
#' @param mainSep the separator for each part of the header
#' @export
#' @return atomstart
atomStart <- function(header, mainSep="|"){
  stopifnot(nchar(header) > 0)
  
  splitHeader <- strsplit(header, mainSep, fixed=TRUE)[[1]]
  startData <- splitHeader[length(splitHeader)]
  atomstart <- as.numeric(substring(startData, 2))
  return(atomstart)
}

#' renumber ATOM
#' 
#' takes an ATOM start offset, and ATOM sequence, and generates a new numbering. Any
#' missing sequence (-) is replaced by zeros.
#' 
#' @param offset the offset to use for renumbering
#' @param sequence the ATOM sequence (AAString)
#' @return new numbering
#' @export
#' @importFrom Biostrings matchPattern AAStringSet AAString
renumATOM <- function(offset, sequence){
  atomRanges <- IRanges(start=1, width=length(sequence))
  atomInsert <- reduce(matchPattern("-", sequence)@ranges)
  
  atomRanges <- disjoin(c(atomRanges, atomInsert))
  rangeType <- rep("real", length(atomRanges))
  suppressWarnings(rangeType[atomRanges %in% atomInsert] <- "deletion")
  
  actualRange <- atomRanges[rangeType %in% "real"]
  outRange <- lapply(actualRange, function(x){x})
  outRange <- do.call(c, outRange)
  outRange2 <- outRange + offset - 1
  
  atomNum <- rep(0, length(sequence))
  atomNum[outRange] <- outRange2
  
  return(atomNum)
}

#' checks that ligands are the same
#' 
#' Runs a check that the ligand locations in the header of the ATOM sequence and
#' the actual ATOM sequence are identical
#' 
#' @param header the header to work with
#' @param atomnum the ATOM numbering
#' @param sequence the ATOM sequence
#' @return logical
#' @export 
checkLigands <- function(ligand, atomnum, sequence){
  ligand <- unique(ligand[, c("loc", "ligand", "lShort")])
  ligand <- ligand[!(ligand$lShort == "-"),]
  
  if (nrow(ligand) == 0){
    return(TRUE)
  }
  
  atomLoc <- which(atomnum %in% ligand$loc)
  
  return(identical(paste(ligand$lShort, sep="", collapse=""), as.character(sequence[atomLoc])))
}

#' check chains in header
#' 
#' we need to check the consistency of the chains listed in the header that Sen provided.
#' Note that we do allow non-amino acid ligand chains to be different
#' 
#' @param header the header to check
#' @param ligandSep the separator of the ligands (default is "|")
#' @param innerSep the separator within a ligand (default is ".")
#' @export
#' @return logical, are the chains all equal
checkChains <- function(header, ligandSep="|", innerSep="."){
  allSame <- FALSE
  
  headerSplit <- strsplit(header, ligandSep, fixed=TRUE)[[1]]
  
  idSplit <- strsplit(headerSplit[1], innerSep, fixed=TRUE)[[1]]
  idChain <- idSplit[2]
  
  ligandInfo <- headerSplit[2:(length(headerSplit)-1)]
  ligChains <- substring(ligandInfo, 1, 1)
  
  ligSplit <- strsplit(ligandInfo, innerSep, fixed=TRUE)
  ligCode <- sapply(ligSplit, function(x){x[2]})
  
  ligAA <- ligCode %in% toupper(AMINO_ACID_CODE)
  
  ligChains <- ligChains[ligAA]
  
  atomChain <- substring(headerSplit[length(headerSplit)], 1, 1)
  
  if (length(unique(c(idChain, ligChains, atomChain))) == 1){
    allSame <- TRUE
  }
  return(allSame)
}

#' read in ATOM numbering file
#' 
#' @param file the file to read
#' @export
#' @return list of vectors
readATOMNum <- function(file){
  atomnum <- scan(file, what=character(), sep="\n")
  
  stopifnot((2 %/% length(atomnum)) == 0)
  natom <- length(atomnum)
  
  headerrow <- seq(1, natom-1, 2)
  numrow <- seq(2, natom, 2)
  headers <- atomnum[headerrow]
  headers <- substring(headers, 2)
  
  numdata <- atomnum[numrow]
  
  numsplit <- strsplit(numdata, ".", fixed=TRUE)
  numnum <- lapply(numsplit, as.numeric)
  names(numnum) <- headers
  
  return(numnum)
}

#' count ligands, aa status and bidentate status
#' 
#' given a list of ligand data frames, determine the number of ligands, whether they are all amino acids,
#' and whether any bidentate ligands exist.
#' 
#' 
#' @param ligands list of ligands
#' @details assume that each entry is a data.frame with "loc", "ligand"
#' @return data.frame with count, AA status (logical), and bidentate status (logical)
#' @export
countLigands <- function(ligands){
  stopifnot(sum(c("loc", "ligand") %in% names(ligands[[1]])) == 2)
  upperAA <- toupper(AMINO_ACID_CODE)
  
  ligandData <- lapply(ligands, function(inLigand){
    nLigand <- nrow(inLigand)
    nAA <- sum(inLigand$ligand %in% upperAA)
    nNonAA <- nLigand - nAA
    isBi <- sum(duplicated(inLigand$loc)) > 0
    allAA <- nAA == nLigand
    
    data.frame(count=nLigand, nAA=nAA, nNonAA=nNonAA, allAA=allAA, isBidentate=isBi)
  })
  
  countData <- do.call(rbind, ligandData)
  countData$id <- names(ligands)
  return(countData)
}

#' generate ligand ranges
#' 
#' given the ligands and their locations, the ATOM numbering, and the ATOM to SEQRES mapping, this function generates
#' an IRanges object containing the range of the ligands involved in metal binding. 
#' 
#' @param ligandLoc data.frame of ligands and positions
#' @param atomnum numeric vector of atom numbering locations
#' @param atom2seqres numeric vector of the SEQRES numbering corresponding to an ATOM location
#' @return IRanges
#' @export
ligandRanges <- function(ligandLoc, atomnum, atom2seqres){
  useLoc <- unique(ligandLoc$loc)
  atomLoc <- which(atomnum %in% useLoc)
  seqLoc <- atom2seqres[atomLoc]
  return(IRanges(start=min(seqLoc), end=max(seqLoc)))
}

#' generate interproscan ranges
#' 
#' @param interproRes the interproresults data.frame
#' @param startCol which column has the beginning of the range (name or index)
#' @param endCol which column has the end of the range (name or index)
#' @param unique_on which column to use to determine unique entries. Most common would be "sigAccession" or "iprAccession"
#' @param require_char which column needs some characters to keep
#' @param asMCOL which columns to put in the mcols field of the generated IRanges object (defaults is "all" other columns)
#' @export
#' @return IRanges
interpro2ranges <- function(interproRes, startCol=7, endCol=8, unique_on = "sigAccession", require_char = "sigAccession", asMCOL="all"){
  allnames <- names(interproRes)
  if (is.numeric(startCol)){
	  startCol <- allnames[startCol]
  }
  if (is.numeric(endCol)){
	  endCol <- allnames[endCol]
  }
  
  uniq_entries <- (!duplicated(interproRes[, unique_on])) & (nchar(interproRes[, require_char]) != 0)
  
  if (asMCOL == "all"){
	  keepNames <- allnames[!(allnames %in% c(startCol, endCol))]
  } else if (is.numeric(asMCOL)){
	  keepNames <- allnames[asMCOL]
  }

  outRanges <- IRanges(start=interproRes[uniq_entries, startCol], end=interproRes[uniq_entries, endCol])
  mcols(outRanges) <- as(interproRes[uniq_entries, keepNames], "DataFrame")

  return(outRanges)
}

#' get ipr annotation results
#' 
#' From a set of Interproscan intervals, we access the mcols data and return any iprAccession and iprDescription data
#' 
#' @param interproRanges the set of ranges
#' @export
#' @return character vector of ipr accessions
getIPRAnnotations <- function(interproRanges){
  mData <- mcols(interproRanges)[, "iprAccession"]
  mData <- mData[!(nchar(mData) == 0)]
  return(unique(mData))
}

#' calculate jaccard similarity
#' 
#' Calculates the jaccard similarity between two character vectors
#' 
#' @param set1 the first set
#' @param set2 the second set
#' @return jaccard similarity (numeric)
#' @export
jaccard <- function(set1, set2){
  length(intersect(set1, set2)) / length(union(set1, set2))
}

#' calculate pairwise jaccard similarities between many
#' 
#' For a set of items, calculate all pairwise similarities using jaccard metric
#' 
#' @param allsets a list of items
#' @export
#' @return matrix
pairwiseJaccard <- function(allsets){
  nSets <- length(allsets)
  
  jaccardMatrix <- matrix(0, nrow=nSets, ncol=nSets)
  allnames <- names(allsets)
  if (!(is.null(allnames))){
    rownames(jaccardMatrix) <- allnames
    colnames(jaccardMatrix) <- allnames
  }
  
  for (iSet in seq(1, nSets)){
    set1 <- allsets[[iSet]]
    for (jSet in (seq(iSet, nSets))){
      set2 <- allsets[[jSet]]
      
      jValue <- jaccard(set1, set2)
      jaccardMatrix[iSet, jSet] <- jaccardMatrix[jSet, iSet] <- jValue
    }
  }
  diag(jaccardMatrix) <- 0
  return(jaccardMatrix)
  
}

#' get pdb and chain id
#' 
#' from a set of headers, we want the PDB and chain part of the header
#' 
#' @param headers a character vector of headers to process
#' @param charsplit the character used for splitting (default is ".")
#' @param pdbIndex which position in the split string holds the pdb id (default is 1)
#' @param chainIndex which position in the split string holds the chain id (default is 2)
#' @export
#' @return data.frame with header, pdb id, and chain id
pdbAndChain <- function(headers, charsplit=".", pdbIndex=1, chainIndex=2){
  splitHeaders <- strsplit(headers, charsplit, fixed=TRUE)
  pdbID <- sapply(splitHeaders, function(x){x[pdbIndex]})
  chainID <- sapply(splitHeaders, function(x){x[chainIndex]})
  
  return(data.frame(header=headers, pdb=pdbID, chain=chainID, stringsAsFactors=FALSE))
  
}

#' check chain diffs
#' 
#' check if there are different chains with the same binding ligands
#' 
#' @param headerChains data.frame with header and chain information
#' @param ligands list of data.frames of ligands from each header
#' @return set of unique headers
#' @export
chainDiffs <- function(headerChains, ligands){
  nChain <- nrow(headerChains)
  
  compMatrix <- matrix(FALSE, nrow=nChain, ncol=nChain)
  for (iChain in seq(1, nChain)){
    for (jChain in seq(iChain, nChain)){
      compMatrix[iChain, jChain] <- identical(ligands[[iChain]], ligands[[jChain]])
    }
  }
  diag(compMatrix) <- FALSE
  nMatch <- apply(compMatrix, 1, sum)
  noMatch <- which(nMatch == 0)
  
  uniqueHeader <- headerChains$header[noMatch]
  return(uniqueHeader)
}

#' check sequences
#' 
#' given a set of sequences that had the same MD5 checksum, check if they have identical binding sites, SEQRES records, 
#' ATOM records, ATOM numbering and ATOM to SEQRES mapping
#' 
#' @param headers set of headers to compare
#' @param ligands the set of ligand binding sites for the supplied headers
#' @param seqres the set of SEQRES sequences
#' @param atom the set of ATOM sequences
#' @param atomnum the set of ATOM numbering
#' @param atom2seq the ATOM 2 SEQRES mapping
#' @return character vector of truly unique headers
#' @export
checkUniqueSequence <- function(headers, ligands, seqres, atom, atomnum, atom2seq){
  nCheck <- length(headers)
  
  compMatrix <- matrix(FALSE, nrow=nCheck, ncol=nCheck)
  
  for (iCheck in seq(1, nCheck)){
    for (jCheck in seq(iCheck, nCheck)){
      compMatrix[iCheck, jCheck] <- identical(ligands[[iCheck]], ligands[[jCheck]]) & 
        identical(as.character(seqres[[iCheck]]), as.character(seqres[[jCheck]])) &
        identical(as.character(atom[[iCheck]]), as.character(atom[[jCheck]])) &
        identical(atomnum[[iCheck]], atomnum[[jCheck]]) & 
        identical(atom2seq[[iCheck]], atom2seq[[jCheck]])
    }
  }
  
  diag(compMatrix) <- FALSE
  nMatch <- apply(compMatrix, 1, sum)
  noMatch <- which(nMatch == 0)
  return(headers[noMatch])
}

#' cluster contributions
#' 
#' given an annotation-gene list, calculate the contribution from each cluster and store it as a matrix of percentages, and store 
#' a character vector
#' 
#' Percentages are calculated for each annotation as:
#' 
#' (countsInCluster / totalClusterSize) / allClusterFractions
#' 
#' @param funcAnnotation list of annotation-gene relationships
#' @param clusterData data.frame of cluster annotations
#' @param clusterID which column to use in the data.frame to identify the cluster
#' @param geneID which column to use that matches the gene ids in the funcAnnotation list
#' @return matrix of percentages, data.frame with a string concatenation of contributions
#' @export
clusterContributions <- function(funcAnnotation, clusterData, clusterID = "cluster", geneID = "long"){
  nAnnotation <- length(funcAnnotation)
  annotNames <- names(funcAnnotation)
  nCluster <- length(unique(clusterData[, clusterID]))
  rleCluster <- rle(sort(clusterData[, clusterID]))
  clusterCounts <- rleCluster$lengths
  names(clusterCounts) <- rleCluster$values
  
  matrixPercentages <- matrix(0, nrow=nAnnotation, ncol=nCluster)
  rownames(matrixPercentages) <- annotNames
  colnames(matrixPercentages) <- names(clusterCounts)
  matrixCounts <- matrixPercentages
  textDescription <- data.frame(annotation = annotNames, clusterCount = "", stringsAsFactors = FALSE)
  rownames(textDescription) <- annotNames
  
  for (iAnnot in annotNames){
    tmpClusters <- clusterData[(clusterData[, geneID] %in% funcAnnotation[[iAnnot]]), clusterID]
    nGene <- length(tmpClusters)
    
    rleSort <- rle(sort(tmpClusters))
    matrixCounts[iAnnot, as.character(rleSort$values)] <- rleSort$lengths
    
    tmpPercentage <- matrixCounts[iAnnot, ] / clusterCounts
    matrixPercentages[iAnnot, ] <- tmpPercentage / sum(tmpPercentage)
  }
  
  textDescription[, "clusterCount"] <- apply(matrixCounts, 1, function(x){paste(x, collapse=",", sep="")})
  
  return(list(percentage = matrixPercentages, count = matrixCounts, text = textDescription))
}


#' cluster contributions 2
#' 
#' given an annotation-gene list, calculate the contribution from each cluster and store it as a matrix of percentages, and store 
#' a character vector
#' 
#' Percentages are calculated for each annotation as:
#' 
#' (countsInCluster / totalClusterSize) / allClusterFractions
#' 
#' @param func_to_site list of annotation-site relationships
#' @param cluster_to_site list of cluster-site relationships
#' @return matrix of percentages, data.frame with a string concatenation of contributions
#' @export
cluster_contributions_2 <- function(func_to_site, cluster_to_site){
  n_func <- length(func_to_site)
  func_names <- names(func_to_site)
  n_cluster <- length(cluster_to_site)
  cluster_counts <- sapply(cluster_to_site, length)
  site_to_cluster <- reverseSplit(cluster_to_site)
  
  matrix_percentages <- matrix(0, nrow=n_func, ncol=n_cluster)
  rownames(matrix_percentages) <- func_names
  colnames(matrix_percentages) <- names(cluster_counts)
  matrix_counts <- matrix_percentages
  text_description <- data.frame(annotation = func_names, cluster_count = "", stringsAsFactors = FALSE)
  rownames(text_description) <- func_names
  
  for (i_func in func_names){
    tmp_cluster <- site_to_cluster[func_to_site[[i_func]]]
    n_gene <- length(tmp_cluster)
    
    rle_tmp <- rle(sort(unlist(tmp_cluster, use.names = FALSE)))
    matrix_counts[i_func, as.character(rle_tmp$values)] <- rle_tmp$lengths
    
    tmp_percent <- matrix_counts[i_func, ] / cluster_counts
    matrix_percentages[i_func, ] <- tmp_percent / sum(tmp_percent)
  }
  
  text_description[, "cluster_count"] <- apply(matrix_counts, 1, function(x){paste(x, collapse=",", sep="")})
  
  return(list(percentage = matrix_percentages, count = matrix_counts, text = text_description))
}


#' cluster contribution similarity
#' 
#' given a matrix of cluster contributions, calculates pearson correlation between the various annotations
#' 
#' @param clusterContributionMatrix matrix of cluster contributions
#' @return data.frame with "from", "to", and pearson similarity
#' @export
contributionSimilarity <- function(clusterContributionMatrix, keepCutoff = 0.9, useAbsolute = FALSE){
  
  annotNames <- rownames(clusterContributionMatrix)
  nAnnotation <- length(annotNames)
  compareGrid <- expand.grid(seq(1, nAnnotation), seq(1, nAnnotation))
  compareGrid <- compareGrid[(compareGrid[,2] > compareGrid[,1]), ]
  compareGrid <- as.matrix(cbind(annotNames[compareGrid[,1]], annotNames[compareGrid[,2]]))
  
  fullSimilarity <- cor(t(clusterContributionMatrix))
  
  if (useAbsolute){
    fullSimilarity <- abs(fullSimilarity)
  }
  
  outEdges <- data.frame(from = compareGrid[,1], to = compareGrid[,2], weight = 0, edgeType = "cluster",
                         stringsAsFactors = FALSE)
  
  outEdges[, "weight"] <- apply(compareGrid, 1, function(x){
    fullSimilarity[x[1], x[2]]
  })
  
  outEdges <- outEdges[(outEdges[, "weight"] >= keepCutoff),]
  return(outEdges)
}



#' all to all similarity
#' 
#' calculate similarity metric between all list elements
#' 
#' @param funcAnnotation list of annotations to calculate overlap for
#' @param similarityType either "overlap" (default) or "jaccard"
#' @return data.frame of edges
#' @export
all2allSimilarity <- function(funcAnnotation, similarityType = "overlap"){
  doComparisons <- genCompareMatrix(names(funcAnnotation))
  
  outEdges <- data.frame(from = doComparisons[,1], to = doComparisons[,2], weight = 0, edgeType = similarityType, stringsAsFactors = FALSE)
  
  similarityFunc <- switch(similarityType,
         overlap = overlapMetric)
  
  outEdges[, "weight"] <- apply(doComparisons, 1, function(x){
    similarityFunc(funcAnnotation[[x[1]]], funcAnnotation[[x[2]]])
  })
  
  return(outEdges)
}

#' cluster proportions
#' 
#' get out the proportions for each cluster for generating random counts
#' 
#' @param clusters a vector of the cluster ids
#' @return named vector of proportions
#' @export
clusterProportions <- function(clusters){
  cP <- rle(sort(clusters))
  cPOut <- cP$lengths
  cPOut <- cPOut / sum(cPOut)
  names(cPOut) <- as.character(cP$values)
  
  return(cPOut)
}

#' expected counts
#' 
#' calculate the expected counts in each cluster based on the total number of genes annotated to each functionality
#' 
#' @param funcAnnotation gene-annotation list
#' @param expectedProportions the proportions in each cluster, should be a matrix of 1 row and however many clusters
#' @export
#' @return matrix, rows for each annotation, columns for each cluster
expectedCounts <- function(funcAnnotation, expectedProportions){
  annotNames <- names(funcAnnotation)
  nAnnot <- sapply(funcAnnotation, length)
  nAnnot <- matrix(nAnnot, length(nAnnot), 1)
  
  expCount <- nAnnot %*% expectedProportions
  rownames(expCount) <- annotNames
  colnames(expCount) <- colnames(expectedProportions)
  
  return(expCount)
}

#' filter function
#' 
#' given a list and a dataframe, keeps only those things in list entries in the data.frame, and only those things in the data.frame
#' in the list
#' 
#' @param listEntries list of vectors
#' @param frameEntry a data.frame
#' @param useFrame which column of the data.frame to filter on
#' @export
#' @return list with filterList and filterFrame
#' @importFrom Biobase reverseSplit
filterListFrame <- function(listEntry, frameEntry, useFrame){
  entryList <- Biobase::reverseSplit(listEntry)
  intersectListFrame <- intersect(names(entryList), frameEntry[, useFrame])
  entryList <- entryList[intersectListFrame]
  
  outList <- Biobase::reverseSplit(entryList)
  outFrame <- frameEntry[(frameEntry[, useFrame] %in% intersectListFrame), ]
  
  return(list(filterList = outList, filterFrame = outFrame))
}

#' p values of differences
#' 
#' for expected and observed count matrices, calculate a p-value for each row
#' 
#' 
#' @param expectedCounts a matrix of expected counts, each row a case, each column an observation
#' @param observedCounts a matrix of observed counts, each row a case, each column an observation
#' @export
#' @return vector p-values
calcChiP <- function(expectedCounts, observedCounts){
  stopifnot(dim(expectedCounts) == dim(observedCounts))
  oeDif <- (observedCounts - expectedCounts) ^ 2 / expectedCounts
  oeSum <- rowSums(oeDif)
  
  pchisq(oeSum, ncol(expectedCounts) - 1, lower.tail = FALSE)
}

#' annotation-cluster co-variances
#' 
#' given the normalized proportions from each cluster, calculate co-variances between each cluster, and then sum across annotations,
#' and normalize to give total co-variance of 1
#' 
#' @param clusterProportions matrix, each row is an annotation, each column is a cluster, total of each row is 1
#' @export
#' @return co-variance matrix
clusterCovariance <- function(clusterProportions){
  nCluster <- ncol(clusterProportions)
  clusterName <- colnames(clusterProportions)
  
  compareClusters <- genCompareMatrix(clusterName)
  
  covMatrix <- matrix(0, nCluster, nCluster)
  rownames(covMatrix) <- colnames(covMatrix) <- clusterName
  
  nComp <- nrow(compareClusters)
  
  for (iComp in seq(1, nComp)){
    irow <- compareClusters[iComp,1]
    jcol <- compareClusters[iComp,2]
    minProps <- apply(clusterProportions[, c(irow, jcol)], 1, min)
    minProps <- minProps^2
    
    covMatrix[irow, jcol] <- covMatrix[jcol, irow] <- sum(minProps)
  }
  
  normMatrix <- covMatrix / max(covMatrix)
  
  return(normMatrix)
}

#' functional distance
#' 
#' calculates the functional distance between clusters based on covariance between clusters using the functional attributes
#' 
#' @param func_to_site list of functional annotations and the sites that contain them
#' @param cluster_to_site list of clusters and the sites that belong to them
#' @param min_site the minimum number of sites to keep an entry in func_to_site and cluster_to_site
#' @export
#' @importFrom Biobase reverseSplit
#' @return list with number of sites and a distance matrix
calculate_functional_distance <- function(func_to_site, cluster_to_site, min_site = 2){
  all_func_site <- unique(unlist(func_to_site))
  all_cluster_site <- unique(unlist(cluster_to_site))
  keep_site <- intersect(all_func_site, all_cluster_site)
  
  func_to_site <- lapply(func_to_site, intersect, keep_site)
  cluster_to_site <- lapply(cluster_to_site, intersect, keep_site)
  
  func_to_site <- (sapply(func_to_site, length) >= min_site) %>% func_to_site[.]
  cluster_to_site <- (sapply(cluster_to_site, length) >= min_site) %>% cluster_to_site[.]
  
  n_site <- length(intersect(unique(unlist(func_to_site)), unique(unlist(cluster_to_site))))
  
  cluster_counts <- cluster_contributions_2(func_to_site, cluster_to_site)
  cluster_variance <- clusterCovariance(cluster_counts$percentage)
  cluster_dist <- as.dist(1 - cluster_variance)
  return(list(dist = cluster_dist, n_site = n_site))
}

#' convert ligands to ranges
#' 
#' given a set of ligands on a chain, if there are more than one ligand on the chain, generate an actual range for the chain
#' 
#' @param in_ligands the set of ligands
#' @export
#' @return IRanges object
convert_ligands_to_ranges <- function(in_ligands){
  n_ligand <- length(in_ligands)
  
  out_range <- in_ligands
  
  if (n_ligand > 1){
    out_range <- IRanges(start = min(start(in_ligands)), end = max(start(in_ligands)))
  }
  
  return(out_range)
}

#' get a unique list of id's
#' 
#' given an \code{IRanges} instance, get a unique list of a particular column from the \code{mcols} \code{DataFrame}.
#' 
#' @param in_ranges the \code{IRanges} instance
#' @param which_col the column to pull from the \code{mcols} \code{DataFrame}
#' @return character vector
#' @export
#' @importFrom S4Vectors mcols
get_unique_mcol <- function(in_ranges, which_col = "iprAccession"){
  unique(mcols(in_ranges)[, which_col])
}

#' collapse multi-chains
#' 
#' given a chain-to-annotation list, and a site-to-chain list, collapse the chain-to-annotations to site-to-annotations
#' 
#' @param chain_to_annotation the list of chains to annotations
#' @param site_to_chain the list of mapping sites to chains
#' @export
#' @return list of site_to_annotation
collapse_chain_to_annotation <- function(chain_to_annotation, site_to_chain){
  all_chains_annotation <- names(chain_to_annotation)
  chain_to_site <- reverseSplit(site_to_chain)
  all_chains_site <- names(chain_to_site)
  
  keep_chains_site <- intersect(all_chains_annotation, all_chains_site)
  chain_to_annotation <- chain_to_annotation[keep_chains_site]
  site_to_chain <- reverseSplit(chain_to_site[keep_chains_site])
  
  lapply(site_to_chain, function(in_sites){
    unique(unlist(chain_to_annotation[keep_chains_site %in% in_sites], use.names = FALSE))
  })
}

#' get ligand to annotation overlap
#' 
#' Given a set of ligand ranges, and a set of annotation ranges, find those annotation ranges that overlap
#' the ligand ranges
#' 
#' Note that the annotation names are used as the default to search on, but everything is only done on the intersection of the two sets of names
#' 
#' @param annotation_ranges a \code{list} of \code{IRanges}, an entry for each annotation
#' @param ligand_ranges a \code{list} of \code{IRanges}, an entry for each ligand
#' @export
#' @return each entry of annotation_ranges subsetted by overlap with the corresponding ligand_ranges
#' @importFrom IRanges subsetByOverlaps
subset_annotation_by_ligand <- function(annotation_ranges, ligand_ranges){
  use_ranges <- intersect(names(annotation_ranges), names(ligand_ranges))
  
  annotation_subset <- lapply(use_ranges, function(range_name){
    subsetByOverlaps(annotation_ranges[[range_name]], ligand_ranges[[range_name]])
  })
  names(annotation_subset) <- use_ranges
  
  return(annotation_subset)
}

#' list distances compare
#' 
#' given two sets of distance lists, do the right comparisons. Assumes that the names in the first will
#' be found in the second
#' 
#' @param dist_a list of distances
#' @param dist_b list of distances
#' @param method the correlation method to use, default is "spearman"
#' @export
#' @importFrom stats cor
#' @return list of correlations, named by names in dist_b
compare_list_distances <- function(dist_a, dist_b, method = "spearman"){
  out_cor <- lapply(names(dist_a), function(in_dist){
    s_dist <- as.vector(dist_a[[in_dist]])
    
    pos_funct <- grep(in_dist, names(dist_b), value = TRUE)
    s_f_cor <- sapply(dist_b[pos_funct], function(in_funct){
      tmp <- cor.test(s_dist, as.vector(in_funct), alternative = "two.sided", method = method)
      c(tmp$estimate, tmp$p.value)
    })
  })
  
  names(out_cor) <- names(dist_a)
  return(out_cor)
}


#' cleanup functional distances
#' 
#' the way that functional distances are generated is not ideal, so this function
#' cleans things up in a way that distances are in the right order and ready
#' for comparison
#' 
#' @param in_dist a distance list from calculate_functional_distance
#' @param n_clust how many clusters were in the distance calculation
#' 
#' @export
#' @return list
cleanup_dist <- function(in_dist, n_clust){
  use_dist <- in_dist$dist
  matrix_dist <- as.matrix(use_dist)
  row_order <- order(as.numeric(rownames(matrix_dist)))
  matrix_dist <- matrix_dist[row_order, row_order]
  
  curr_entries <- as.numeric(rownames(matrix_dist))
  
  true_clust <- seq(1, n_clust)
  
  if (length(curr_entries) != length(true_clust)){
    zero_matrix <- matrix(0, nrow = length(true_clust), ncol = length(true_clust))
    rownames(zero_matrix) <- colnames(zero_matrix) <- as.character(true_clust)
    
    for (i in rownames(matrix_dist)){
      for (j in rownames(matrix_dist)){
        zero_matrix[i, j] <- matrix_dist[i, j]
      }
    }
    matrix_dist <- zero_matrix
  }
  
  in_dist$dist <- as.dist(matrix_dist)
  in_dist
}
